@@ -53,6 +53,7 @@ CRON_month = "month"
 CRON_dayOfWeek = "dayOfWeek"
 
 REGEX__dmbs_object_name = r'^[0-9a-zA-Z_]{1,63}$'
+REGEX__dmbs_procedure_name = r'^[0-9a-zA-Z_$.]{1,63}$'
 
 SEPARATOR__comma_space = ", "
 SEPARATOR__comma = ","
@@ -105,6 +106,7 @@ CONFIG_KEY__server = "SERVER"
 CONFIG_KEY_SERVER__port = "port"
 
 JWT_PURPOSE__oauth = "oauth"
+JWT_PURPOSE__oidc = "oidc"
 JWT_PURPOSE__pre_auth = "pre_auth"
 JWT_PURPOSE__connection = "connection"
 
@@ -180,5 +182,5 @@ ROLE__postgres = "postgres"
 
 PROTOCOL__postgres = "postgresql://"
 
-VERSION = "4.25.7"
+VERSION = "4.26.1"
 
